/***************************************************************************
 *   Copyright (C) 2009 by Stanson <stanson@pangolin.ru>                   *
 *   GPL v3 code, read COPYING file                                        *
 **************************************************************************/
#define extern
#include <pic18fregs.h>
#undef extern
#include "bootloader.h"
#include "descr.h"
#include "config.h"

data at (0x400) BDT ep0Bo;         //Endpoint #0 BD Out
data at (0x404) BDT ep0Bi;         //Endpoint #0 BD In
data at (0x408) BDT ep1Bo;         //Endpoint #1 BD Out
data at (0x40C) BDT ep1Bi;         //Endpoint #1 BD In

#pragma udata bank4 SetupPkt
CTRL_TRF_SETUP   SetupPkt;
#pragma udata bank4 dataPacket
BOOT_DATA_PACKET dataPacket;
#pragma udata bank4 ctrl_trf_data
byte ctrl_trf_data[EP0_BUFF_SIZE];

#pragma udata bank4 trf_state
data byte trf_state;
#pragma udata bank4 big_counter
data word big_counter;
#pragma udata bank4 ctrl_trf_state
data byte ctrl_trf_state;            // Control Transfer State
#pragma udata bank4 ctrl_trf_count
data word ctrl_trf_count;            // Data counter
#pragma udata bank4 ctrl_trf_mem
data byte ctrl_trf_mem;

#pragma udata bank4 pSrc
data POINTER pSrc;                   // Data source pointer
#pragma udata bank4 pDst
data POINTER pDst;                   // Data destination pointer

#pragma udata bank4 usb_device_state
data byte usb_device_state;          // Device States: DETACHED, ATTACHED, ...
#pragma udata bank4 usb_active_cfg
data byte usb_active_cfg;            // Value of current configuration
#pragma udata bank4 usb_new_addr
data byte usb_new_addr;              // Address sent by host


#pragma code _reset 0x000000
void _reset (void) _naked
{
    _asm
    GOTO _main
    _endasm;
}

#pragma code _high_ISR 0x000008
void _high_ISR (void) _naked
{
    _asm
    GOTO RM_HIGH_INTERRUPT_VECTOR
    _endasm;
}

#pragma code _low_ISR 0x000018
void _low_ISR (void) _naked
{
    _asm
    GOTO RM_LOW_INTERRUPT_VECTOR
    _endasm;
}

void USBPrepareForNextSetupTrf(void)
{
    ctrl_trf_state = WAIT_SETUP;          // See usbctrltrf.h
    ep0Bo.Cnt  = EP0_BUFF_SIZE;           //  
    ep0Bo.ADR  = (data byte*)&SetupPkt;   //  SetupPkt
    ep0Bo.Stat = _USIE | _DAT0 | _DTSEN;  // EP0 buff dsc init, see usbmmap.h
    ep0Bi.Stat = _UCPU;                   // EP0 IN buffer initialization
}

void USBCtrlTrfTxService(void)
{    
    byte byte_to_send;
    
    /*
     * First, have to figure out how many byte of data to send.
     */
    if(ctrl_trf_count < EP0_BUFF_SIZE)
        byte_to_send = ctrl_trf_count;
    else
        byte_to_send = EP0_BUFF_SIZE;
    
    ep0Bi.Cnt = byte_to_send;
    
    /*
     * Subtract the number of bytes just about to be sent from the total.
     */
    ctrl_trf_count -= byte_to_send;
    
    _asm LFSR 0, _ctrl_trf_data _endasm;

    if(ctrl_trf_mem == _ROM)
    {
        /*   ROM */
        TBLPTRL = pSrc.bLow;
        TBLPTRH = pSrc.bHigh;
        TBLPTRU = pSrc.bUpper;
        while(byte_to_send)
        {
            _asm TBLRD*+ _endasm;
	    POSTINC0 = TABLAT;
            byte_to_send--;
	}
	/*   */
        pSrc.bLow = TBLPTRL;
        pSrc.bHigh = TBLPTRH;
        pSrc.bUpper = TBLPTRU;
    }
    else
    {
        /*   RAM */
        FSR2L = pSrc.bLow;
        FSR2H = pSrc.bHigh;
        while(byte_to_send)
        {
            POSTINC0 = POSTINC2;
            byte_to_send--;
	}
	/*   */
        pSrc.bLow = FSR2L;
        pSrc.bHigh = FSR2H;
    }
}//end USBCtrlTrfTxService

//void USBCtrlTrfRxService(void)
//{
//    byte byte_to_read;

//    byte_to_read = ep0Bo.Cnt;
    
    /*
     * Accumulate total number of bytes read
     */
//    ctrl_trf_count += byte_to_read;
    
//    FSR0L = pDst.bLow;
//    FSR0H = pDst.bHigh;
//    _asm LFSR 2, _ctrl_trf_data _endasm;

//    while(byte_to_read)
//    {
//        POSTINC0 = POSTINC2;
//        byte_to_read--;
//    }
//}//end USBCtrlTrfRxService

void USBCtrlEPServiceComplete(void)
{
    if(SetupPkt.bmRequestType & 0x80) // DEV_TO_HOST
    {
        /* Control Read */
        ctrl_trf_state = CTRL_TRF_TX;

        if(SetupPkt.wLength < ctrl_trf_count)
            ctrl_trf_count = SetupPkt.wLength;

        /*        */
        USBCtrlTrfTxService();

        /* Data stage -    */
        ep0Bi.ADR = (data byte*)&ctrl_trf_data;

        /* Status stage -   OUT(1) */
        ep0Bo.Cnt = EP0_BUFF_SIZE;
        ep0Bo.ADR = (data byte*)&SetupPkt;            
        ep0Bo.Stat = _USIE;           // Note: DTSEN is 0!
    }
    else // HOST_TO_DEV
    {
        /* Control Write  No-data Control */
        ctrl_trf_state = CTRL_TRF_RX;

        /* Data stage -     */
        ep0Bo.Cnt = EP0_BUFF_SIZE;
        ep0Bo.ADR = (data byte*)&ctrl_trf_data;
        ep0Bo.Stat = _USIE | _DAT1 | _DTSEN;

        /* Status stage -   IN(1) */
        ep0Bi.Cnt = 0;
    }
}

void USBCtrlStall(void)
{
    /*  STALL     SETUP */
    ep0Bo.Cnt = EP0_BUFF_SIZE;
    ep0Bo.ADR = (data byte*)&SetupPkt;
    ep0Bo.Stat = _USIE | _BSTALL;

    ep0Bi.Stat = _USIE | _BSTALL;
}

void USBCheckStdRequest(void)
{   
    ctrl_trf_count = 0;

    if(SetupPkt.bmRequestType & 0x60) goto Stall;

    switch(SetupPkt.bRequest)
    {
        case SET_ADR:
            usb_new_addr     = (byte)SetupPkt.wValue;   //   
            usb_device_state = ADR_PENDING_STATE;       // Update state only
	    break;
        case GET_DSC:
	    switch(SetupPkt.bDscType)
	    {
                case DSC_DEV:
                    pSrc.bRom = (code byte*)&device_dsc;
                    ctrl_trf_count = sizeof(device_dsc);
		    break;
                case DSC_CFG:
                    pSrc.bRom = (code byte*)&cfg01;
                    ctrl_trf_count = sizeof(cfg01);
		    break;
                case DSC_STR:
		    switch(SetupPkt.bDscIndex)
		    {
		        case 1:
	                    pSrc.bRom = (code byte*)&sd001;
	                    ctrl_trf_count = sizeof(sd001);
			    break;
		        case 2:
	                    pSrc.bRom = (code byte*)&sd002;
	                    ctrl_trf_count = sizeof(sd002);
			    break;
		        case 3:
	                    pSrc.bRom = (code byte*)&sd003;
	                    ctrl_trf_count = sizeof(sd003);
			    break;
		        default:
	                    pSrc.bRom = (code byte*)&sd000;
	                    ctrl_trf_count = sizeof(sd000);
			    break;
		    }
		    break;
                default: //   
		    goto Stall;
	    }
            ctrl_trf_mem = _ROM;               // Set memory type
            break;
        case SET_CFG:
            usb_active_cfg = (byte)SetupPkt.wValue;

            if(usb_active_cfg == 0)
	    {
                usb_device_state = ADDRESS_STATE;
	    }
            else
            {
                usb_device_state = CONFIGURED_STATE;

                trf_state = WAIT_FOR_CMD;
                UEP1 = EP_OUT_IN | HSHK_EN;         // Enable 2 data pipes

                ep1Bo.Cnt  = sizeof(dataPacket);      // buffer size
                ep1Bo.ADR  = (data byte*)&dataPacket; // buffer address
                ep1Bo.Stat = _USIE |_DAT0 |_DTSEN;    // status

                ep1Bi.ADR  = (data byte*)&dataPacket; // buffer address
                ep1Bi.Stat = _UCPU | _DAT1;           // status
            }
            break;
        case GET_CFG:
            pSrc.bRam = (data byte*)&usb_active_cfg;    // Set Source
            ctrl_trf_mem = _RAM;                        // Set memory type
            ctrl_trf_count = 1;                         // Set data count
            break;
        default: //  
	    goto Stall;
    }
    USBCtrlEPServiceComplete();
    return;
Stall:
    USBCtrlStall();
    return;
}

void USBCtrlEPService(void)
{   
    if(USTAT == EP00_OUT)
    {
        if((ep0Bo.Stat & TOKEN_TYPE_MASK) == SETUP_TOKEN) // EP0 SETUP
	{
            USBCheckStdRequest();
            ep0Bi.Stat = _USIE | _DAT1 | _DTSEN;
            UCONbits.PKTDIS = 0;
	}
        else                                        // EP0 OUT
	{
	    if(ctrl_trf_state == CTRL_TRF_RX)
	    {
//        	USBCtrlTrfRxService();
        
	        if(ep0Bo.Stat & _DTSMASK) // DTS == 1
	            ep0Bo.Stat = _USIE | _DAT0 | _DTSEN;
	        else
	            ep0Bo.Stat = _USIE | _DAT1 | _DTSEN;
	    }
	    else
	    {
	        USBPrepareForNextSetupTrf();
	    }
        }
    }
    else if(USTAT == EP00_IN)                       // EP0 IN
    {
	if(usb_device_state == ADR_PENDING_STATE)
        {
            UADDR = usb_new_addr;
            if(UADDR)
	    {
                usb_device_state = ADDRESS_STATE;
            }
	    else
	    {
                usb_device_state = DEFAULT_STATE;
	    }
        }

        if(ctrl_trf_state == CTRL_TRF_TX)
        {
            USBCtrlTrfTxService();
        
            if(ep0Bi.Stat & _DTSMASK) // DTS == 1
                ep0Bi.Stat = _USIE | _DAT0 | _DTSEN;
            else
                ep0Bi.Stat = _USIE | _DAT1 | _DTSEN;
        }
        else
	{
           USBPrepareForNextSetupTrf();
	}
    }
}//end USBCtrlEPService

void USBDriverService(void)
{   
byte flags = UIR & UIE; 

    if(flags & (1<<2)) // ACTIVIF
    {
        UCONbits.SUSPND = 0;
        UIEbits.ACTVIE = 0;
        UIRbits.ACTVIF = 0;
        UIRbits.IDLEIF = 0;
        UIEbits.IDLEIE = 1;
    }
            
    if(flags & (1<<0)) // URSTIF
    {
        while(UIRbits.TRNIF)            // Flush any pending transactions
            UIRbits.TRNIF = 0;
        UIR   = 0x00;                   // Clears all USB interrupts
        UIE   = 0x7B;                   // Enable all interrupts except ACTVIE
        UADDR = 0x00;                   // Reset to default address
        UEP0  = EP_CTRL | HSHK_EN;      // Init EP0 as a Ctrl EP
        USBPrepareForNextSetupTrf();    //  BDT  ɣ SETUP
        UCONbits.PKTDIS = 0;            // Make sure packet processing is enabled
        usb_active_cfg = 0;             // Clear active configuration
        usb_device_state = DEFAULT_STATE;
    }
    
    if(flags & (1<<4)) // IDLEIF
    {
        UIEbits.ACTVIE = 1;                    // Enable bus activity interrupt
        UIRbits.IDLEIF  = 0;
        UCONbits.SUSPND = 1;                    // Put USB module in power conserve
                                                // mode, SIE clock inactive
        /* Now, go into power saving */
//        PIE2bits.USBIE  = 1;                     // Set wakeup source
//        Sleep();
//        PIR2bits.USBIF  = 0;
    }
    
    if(flags & (1<<5)) // STALLIF
    {
        if(UEP0bits.EPSTALL)
        {
            USBPrepareForNextSetupTrf();        // Firmware Work-Around
            UEP0bits.EPSTALL = 0;
        }
        UIRbits.STALLIF = 0;
    }

    if(usb_device_state < DEFAULT_STATE) return;

    if(flags & (1<<3)) // TRNIF
    {
        USBCtrlEPService();
        UIRbits.TRNIF = 0;
    }
}

void StartWrite(void)
{
    EECON2 = 0x55;
    EECON2 = 0xAA;
    EECON1bits.WR = 1;
}

void ReadProgMem(void)
{
    byte i;

    TBLPTRL = dataPacket.ADR.low;
    TBLPTRH = dataPacket.ADR.high;
    TBLPTRU = dataPacket.ADR.upper;
    _asm LFSR 0, (_dataPacket + 5) _endasm;
    
    for (i = 0; i < dataPacket.len; i++)
    {
	_asm TBLRD*+ _endasm;
	POSTINC0 = TABLAT;
    }
}

void WriteProgMem(void)
{
    byte i;

    EECON1  = 0x84;                  // Setup writes: EEPGD=1,WREN=1
    TBLPTRL = dataPacket.ADR.low;
    TBLPTRH = dataPacket.ADR.high;
    TBLPTRU = dataPacket.ADR.upper;
    TBLPTRL &= 0xE0;                // Force 32-byte boundary
    _asm LFSR 0, (_dataPacket + 5) _endasm;

    for (i = 0; i < dataPacket.len; i++)
    {
        TABLAT = POSTINC0;
	_asm TBLWT*+ _endasm;
    }
    _asm TBLRD*- _endasm;
    TBLPTRL &= 0xE0;                // Force 32-byte boundary
    StartWrite();
}

void EraseProgMem(void)
{
    EECON1  = 0x94;                 //Setup writes: EEPGD=1,FREE=1,WREN=1
    TBLPTRL = dataPacket.ADR.low;
    TBLPTRH = dataPacket.ADR.high;
    TBLPTRU = dataPacket.ADR.upper;
    TBLPTRL &= 0xC0;                // Force 64-byte boundary

    StartWrite();
}

void ReadEE(void)
{
    byte i;

    EECON1 = 0x00;
    EEADR  = dataPacket.ADR.low;
    _asm LFSR 0, (_dataPacket + 5) _endasm;

    for(i = 0; i < dataPacket.len; i++)
    {
        EECON1bits.RD = 1;
	POSTINC0 = EEDATA;
	EEADR++;
    }
}

void WriteEE(void)
{
    byte i;

    EEADR = dataPacket.ADR.low;
    _asm LFSR 0, (_dataPacket + 5) _endasm;

    for(i = 0; i < dataPacket.len; i++)
    {
        EEDATA = POSTINC0;
        EECON1 = 0x04;          //Setup writes: EEPGD=0,WREN=1
        StartWrite();
        while(EECON1bits.WR);   //Wait till WR bit is clear
	EEADR++;
    }
}

//WriteConfig is different from WriteProgMem b/c it can write a byte
void WriteConfig(void)
{
    byte i;

    EECON1  = 0xC4;          //Setup writes: EEPGD=1,CFGS=1,WREN=1
    TBLPTRL = dataPacket.ADR.low;
    TBLPTRH = dataPacket.ADR.high;
    TBLPTRU = dataPacket.ADR.upper;
    _asm LFSR 0, (_dataPacket + 5) _endasm;

    for (i = 0; i < dataPacket.len; i++)
    {
        TABLAT = POSTINC0;
	__asm TBLWT*+ __endasm;
        StartWrite();
    }
}

void BootService(void)
{
    byte i;

    if((usb_device_state < CONFIGURED_STATE) || (UCONbits.SUSPND)) return;

    if(trf_state == SENDING_RESP)
    {
        if(!(ep1Bi.Stat & _USIE))
        { //  .    
            ep1Bo.Cnt = sizeof(dataPacket);

	    if(ep1Bo.Stat & _DTSMASK)
	        ep1Bo.Stat = _USIE | _DAT0 | _DTSEN;
            else
	        ep1Bo.Stat = _USIE | _DAT1 | _DTSEN;

            trf_state = WAIT_FOR_CMD;
        }
        return;
    }

    if(!(ep1Bo.Stat & _USIE))
    { //  ,  
        i = 0x01;
        switch(dataPacket.CMD)
        {
            case READ_VERSION:
                dataPacket._byte[2] = MINOR_VERSION;
                dataPacket._byte[3] = MAJOR_VERSION;
                i = 0x04;
                break;

            case READ_FLASH:
            case READ_CONFIG:
                ReadProgMem();
                i = dataPacket.len + 0x05;
                break;

            case WRITE_FLASH:
                WriteProgMem();
                break;

            case ERASE_FLASH:
                EraseProgMem();
                break;

            case READ_EEDATA:
                ReadEE();
                i = dataPacket.len + 0x05;
                break;

            case WRITE_EEDATA:
                WriteEE();
                break;

            case WRITE_CONFIG:
                WriteConfig();
                break;

            case RESET:
                //When resetting, make sure to drop the device off the bus
                //for a period of time. Helps when the device is suspended.
                UCONbits.USBEN = 0;
                big_counter = 0;
                while(--big_counter);

                Reset();
                break;

            default:
	        i = 0; // Unknown command
                break;
        }
        trf_state = SENDING_RESP;
        if(i != 0)
        { //   
            ep1Bi.Cnt = i;

	    if(ep1Bi.Stat & _DTSMASK)
	        ep1Bi.Stat = _USIE | _DAT0 | _DTSEN;
            else
	        ep1Bi.Stat = _USIE | _DAT1 | _DTSEN;
        }
    }
}

void main(void)
{
    /*
     *      .   RE3
     * RE3  .    MCLRE  (CONFIG3H:7) 
     */
    if(PORTEbits.RE3)
    {
        _asm GOTO RM_RESET_VECTOR _endasm;
    }
    
    /*   */
    ADCON1 |= 0x0F;
    /*   */
    _asm LFSR 1, 0x2FF _endasm;
    /*  USB */
    ep0Bo.Stat = _UCPU;             // EP0 OUT BDT
    ep0Bi.Stat = _UCPU;             // EP0 IN  BDT
    UCFG = _PUEN | _TRINT | _FS | _PPBM0;
    UCON = 0x08;                    // Enable USBEN only
    while(UCONbits.SE0);            // Blocking loop
    UIR = 0;                        // Clear all USB interrupts
    UIE = 0x11;                     // Unmask RESET & IDLE interrupts only

    usb_device_state = POWERED_STATE;

    while(1)
    {
        USBDriverService();
        BootService();
    }
}


//#pragma code user RM_RESET_VECTOR
